"""
The AgentsAPI class provides methods to register and interact with AI Agents on Nevermined.
"""

import requests
from typing import Dict, Any, List, Optional
from payments_py.common.payments_error import PaymentsError
from payments_py.common.types import (
    PaymentOptions,
    AgentMetadata,
    AgentAPIAttributes,
    PlanMetadata,
    PlanPriceConfig,
    PlanCreditsConfig,
    PaginationOptions,
)
from payments_py.api.base_payments import BasePaymentsAPI
from payments_py.api.nvm_api import (
    API_URL_REGISTER_AGENT,
    API_URL_GET_AGENT,
    API_URL_ADD_PLAN_AGENT,
    API_URL_REMOVE_PLAN_AGENT,
    API_URL_GET_AGENT_ACCESS_TOKEN,
    API_URL_REGISTER_AGENTS_AND_PLAN,
    API_URL_GET_AGENT_PLANS,
    API_URL_UPDATE_AGENT,
)


class AgentsAPI(BasePaymentsAPI):
    """
    The AgentsAPI class provides methods to register and interact with AI Agents on Nevermined.
    """

    @classmethod
    def get_instance(cls, options: PaymentOptions) -> "AgentsAPI":
        """
        Get a singleton instance of the AgentsAPI class.

        Args:
            options: The options to initialize the payments class

        Returns:
            The instance of the AgentsAPI class
        """
        return cls(options)

    def register_agent(
        self,
        agent_metadata: AgentMetadata,
        agent_api: AgentAPIAttributes,
        payment_plans: List[str],
    ) -> Dict[str, str]:
        """
        Registers a new AI Agent on Nevermined.
        The agent must be associated to one or multiple Payment Plans. Users that are subscribers of a payment plan can query the agent.
        Depending on the Payment Plan and the configuration of the agent, the usage of the agent/service will consume credits.
        When the plan expires (because the time is over or the credits are consumed), the user needs to renew the plan to continue using the agent.

        Args:
            agent_metadata: Agent metadata
            agent_api: Agent API attributes
            payment_plans: The list of payment plans giving access to the agent

        Returns:
            The unique identifier of the newly created agent (Agent Id)

        Raises:
            PaymentsError: If registration fails
        """
        body = {
            "metadataAttributes": self.pydantic_to_dict(agent_metadata),
            "agentApiAttributes": self.pydantic_to_dict(agent_api),
            "plans": payment_plans,
        }

        options = self.get_backend_http_options("POST", body)
        url = f"{self.environment.backend}{API_URL_REGISTER_AGENT}"

        response = requests.post(url, **options)
        if not response.ok:
            try:
                error = response.json()
            except Exception:
                error = {"message": response.text, "code": response.status_code}
            raise PaymentsError.from_backend("Unable to register agent", error)
        agent_data = response.json()
        return {"agentId": agent_data["agentId"]}

    def register_agent_and_plan(
        self,
        agent_metadata: AgentMetadata,
        agent_api: AgentAPIAttributes,
        plan_metadata: PlanMetadata,
        price_config: PlanPriceConfig,
        credits_config: PlanCreditsConfig,
    ) -> Dict[str, str]:
        """
        Registers a new AI Agent and a Payment Plan associated to this new agent.
        Depending on the Payment Plan and the configuration of the agent, the usage of the agent/service will consume credits.
        When the plan expires (because the time is over or the credits are consumed), the user needs to renew the plan to continue using the agent.

        Args:
            agent_metadata: Agent metadata
            agent_api: Agent API attributes
            plan_metadata: Plan metadata
            price_config: Plan price configuration
            credits_config: Plan credits configuration

        Returns:
            Dictionary containing agentId, planId, and txHash

        Raises:
            PaymentsError: If registration fails
        """
        body = {
            "plan": {
                "metadataAttributes": self.pydantic_to_dict(plan_metadata),
                "priceConfig": self.pydantic_to_dict(price_config),
                "creditsConfig": self.pydantic_to_dict(credits_config),
            },
            "agent": {
                "metadataAttributes": self.pydantic_to_dict(agent_metadata),
                "agentApiAttributes": self.pydantic_to_dict(agent_api),
            },
        }

        options = self.get_backend_http_options("POST", body)
        url = f"{self.environment.backend}{API_URL_REGISTER_AGENTS_AND_PLAN}"

        response = requests.post(url, **options)
        if not response.ok:
            try:
                error = response.json()
            except Exception:
                error = {"message": response.text, "code": response.status_code}
            raise PaymentsError.from_backend("Unable to register agent & plan", error)
        result = response.json()

        return {
            "agentId": result["data"]["agentId"],
            "planId": result["data"]["planId"],
            "txHash": result["txHash"],
        }

    def get_agent(self, agent_id: str) -> Dict[str, Any]:
        """
        Gets the metadata for a given Agent identifier.

        Args:
            agent_id: The unique identifier of the agent

        Returns:
            The agent's metadata

        Raises:
            PaymentsError: If the agent is not found
        """
        url = f"{self.environment.backend}{API_URL_GET_AGENT.format(agent_id=agent_id)}"
        response = requests.get(url)
        if not response.ok:
            try:
                error = response.json()
            except Exception:
                error = {"message": response.text, "code": response.status_code}
            raise PaymentsError.from_backend("Agent not found", error)
        return response.json()

    def get_agent_plans(
        self, agent_id: str, pagination: Optional[PaginationOptions] = None
    ) -> Dict[str, Any]:
        """
        Gets the list of plans that can be ordered to get access to an agent.

        Args:
            agent_id: The unique identifier of the agent
            pagination: Optional pagination options to control the number of results returned

        Returns:
            The list of all different plans giving access to the agent

        Raises:
            PaymentsError: If the agent is not found
        """
        if pagination is None:
            pagination = PaginationOptions()

        url = f"{self.environment.backend}{API_URL_GET_AGENT_PLANS.format(agent_id=agent_id)}"
        params = {
            "page": pagination.page,
            "offset": pagination.offset,
        }
        response = requests.get(url, params=params)
        if not response.ok:
            try:
                error = response.json()
            except Exception:
                error = {"message": response.text, "code": response.status_code}
            raise PaymentsError.from_backend("Unable to get agent plans", error)
        return response.json()

    def add_plan_to_agent(self, plan_id: str, agent_id: str) -> Dict[str, Any]:
        """
        Add a plan to an agent.

        Args:
            plan_id: The unique identifier of the plan
            agent_id: The unique identifier of the agent

        Returns:
            The result of the operation

        Raises:
            PaymentsError: If unable to add plan to agent
        """
        options = self.get_backend_http_options("POST")
        url = f"{self.environment.backend}{API_URL_ADD_PLAN_AGENT.format(agent_id=agent_id, plan_id=plan_id)}"

        response = requests.post(url, **options)
        if not response.ok:
            try:
                error = response.json()
            except Exception:
                error = {"message": response.text, "code": response.status_code}
            raise PaymentsError.from_backend("Unable to add plan to agent", error)
        return response.json()

    def remove_plan_from_agent(self, plan_id: str, agent_id: str) -> Dict[str, Any]:
        """
        Remove a plan from an agent.

        Args:
            plan_id: The unique identifier of the plan
            agent_id: The unique identifier of the agent

        Returns:
            The result of the operation

        Raises:
            PaymentsError: If unable to remove plan from agent
        """
        url = f"{self.environment.backend}{API_URL_REMOVE_PLAN_AGENT.format(agent_id=agent_id, plan_id=plan_id)}"
        options = self.get_backend_http_options("DELETE")

        response = requests.delete(url, **options)
        if not response.ok:
            try:
                error = response.json()
            except Exception:
                error = {"message": response.text, "code": response.status_code}
            raise PaymentsError.from_backend("Unable to remove plan from agent", error)
        return response.json()

    def get_agent_access_token(self, plan_id: str, agent_id: str) -> Dict[str, Any]:
        """
        Get an access token for an agent.

        Args:
            plan_id: The unique identifier of the plan
            agent_id: The unique identifier of the agent

        Returns:
            The access token information

        Raises:
            PaymentsError: If unable to get access token
        """
        url = f"{self.environment.backend}{API_URL_GET_AGENT_ACCESS_TOKEN.format(plan_id=plan_id, agent_id=agent_id)}"
        options = self.get_backend_http_options("GET")

        response = requests.get(url, **options)
        if not response.ok:
            try:
                error = response.json()
            except Exception:
                error = {"message": response.text, "code": response.status_code}
            raise PaymentsError.from_backend("Unable to get agent access token", error)
        return response.json()

    def update_agent_metadata(
        self,
        agent_id: str,
        agent_metadata: AgentMetadata,
        agent_api: AgentAPIAttributes,
    ) -> Dict[str, Any]:
        """
        Updates the metadata and API attributes of an existing AI Agent.

        Args:
            agent_id: The unique identifier of the agent
            agent_metadata: The new metadata attributes for the agent
            agent_api: The new API attributes for the agent

        Returns:
            The result of the update operation

        Raises:
            PaymentsError: If the agent is not found or if the update fails
        """
        body = {
            "metadataAttributes": self.pydantic_to_dict(agent_metadata),
            "agentApiAttributes": self.pydantic_to_dict(agent_api),
        }
        url = f"{self.environment.backend}{API_URL_UPDATE_AGENT.format(agent_id=agent_id)}"
        options = self.get_backend_http_options("PUT", body)
        response = requests.put(url, **options)
        if not response.ok:
            try:
                error = response.json()
            except Exception:
                error = {"message": response.text, "code": response.status_code}
            raise PaymentsError.from_backend("Error updating agent", error)
        return response.json()



"""
Nevermined Payments API endpoints and backend client
"""

import requests
from typing import Optional, Dict, Any

# Plan endpoints
API_URL_REGISTER_PLAN = "/api/v1/protocol/plans"
API_URL_GET_PLAN = "/api/v1/protocol/plans/{plan_id}"
API_URL_PLAN_BALANCE = "/api/v1/protocol/plans/{plan_id}/balance/{holder_address}"
API_URL_ORDER_PLAN = "/api/v1/protocol/plans/{plan_id}/order"
API_URL_MINT_PLAN = "/api/v1/protocol/plans/mint"
API_URL_MINT_EXPIRABLE_PLAN = "/api/v1/protocol/plans/mintExpirable"
API_URL_BURN_PLAN = "/api/v1/protocol/plans/burn"
API_URL_GET_PLAN_AGENTS = "/api/v1/protocol/plans/{plan_id}/agents"
API_URL_REDEEM_PLAN = "/api/v1/protocol/plans/redeem"

# Agent endpoints
API_URL_REGISTER_AGENT = "/api/v1/protocol/agents"
API_URL_GET_AGENT = "/api/v1/protocol/agents/{agent_id}"
API_URL_SEARCH_AGENTS = "/api/v1/protocol/agents/search"
API_URL_ADD_PLAN_AGENT = "/api/v1/protocol/agents/{agent_id}/plan/{plan_id}"
API_URL_REMOVE_PLAN_AGENT = "/api/v1/protocol/agents/{agent_id}/plan/{plan_id}"
API_URL_INITIALIZE_AGENT = "/api/v1/protocol/agents/initialize/{agent_id}"
API_URL_TRACK_AGENT_SUB_TASK = "/api/v1/protocol/agent-sub-tasks"
API_URL_REGISTER_AGENTS_AND_PLAN = "/api/v1/protocol/agents/plans"
API_URL_GET_AGENT_PLANS = "/api/v1/protocol/agents/{agent_id}/plans"
API_URL_UPDATE_AGENT = "/api/v1/protocol/agents/{agent_id}"

# Token endpoints
API_URL_GET_AGENT_ACCESS_TOKEN = "/api/v1/protocol/token/{plan_id}/{agent_id}"
API_URL_VALIDATE_AGENT_ACCESS_TOKEN = "/api/v1/protocol/token/validate/{agent_id}"

# Stripe endpoints
API_URL_STRIPE_CHECKOUT = "/api/v1/stripe/checkout"


class HTTPRequestOptions:
    """
    HTTP request options for Nevermined Payments API.

    :param send_through_proxy: Whether to send the request through the proxy (default True)
    :param proxy_host: Proxy host to use (optional)
    :param headers: Additional headers for the request (optional)
    :param access_token: Access token for authorization (optional)
    """

    def __init__(
        self,
        send_through_proxy: bool = True,
        proxy_host: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
        access_token: Optional[str] = None,
    ):
        self.send_through_proxy = send_through_proxy
        self.proxy_host = proxy_host
        self.headers = headers or {}
        self.access_token = access_token


class AbstractHTTPClient:
    """
    Abstract HTTP client for Nevermined API requests.
    Equivalent to the TypeScript AbstractHTTPClient class.
    """

    def __init__(self):
        """
        Initialize the HTTP client.
        """

    def parse_url(self, url_requested: str, req_options: HTTPRequestOptions) -> str:
        """
        Parse URL for request, using proxy if needed.
        """
        if req_options.send_through_proxy:
            if not req_options.proxy_host:
                raise ValueError("Proxy host is required when sendThroughProxy is true")
            from urllib.parse import urlparse

            proxy_origin = urlparse(req_options.proxy_host).netloc
            return f"https://{proxy_origin}{url_requested}"
        else:
            return url_requested

    def parse_headers(
        self, request_headers: Dict[str, str], access_token: Optional[str] = None
    ) -> Dict[str, str]:
        """
        Parse headers for request, including access token if provided.
        """
        headers = {}
        if access_token:
            headers["Authorization"] = f"Bearer {access_token}"
        headers.update(request_headers or {})
        return headers

    def request(
        self,
        method: str,
        url: str,
        data: Any = None,
        req_options: Optional[HTTPRequestOptions] = None,
    ):
        """
        Make an HTTP request.
        """
        if req_options is None:
            req_options = HTTPRequestOptions(send_through_proxy=False)

        full_url = self.parse_url(url, req_options)
        headers = self.parse_headers(req_options.headers, req_options.access_token)

        try:
            if method.upper() == "GET":
                response = requests.get(full_url, headers=headers)
            elif method.upper() == "POST":
                response = requests.post(full_url, json=data, headers=headers)
            elif method.upper() == "PUT":
                response = requests.put(full_url, json=data, headers=headers)
            elif method.upper() == "DELETE":
                response = requests.delete(full_url, json=data, headers=headers)
            elif method.upper() == "PATCH":
                response = requests.patch(full_url, json=data, headers=headers)
            else:
                raise ValueError(f"Unsupported HTTP method: {method}")

            response.raise_for_status()
            return response

        except requests.HTTPError as err:
            try:
                message = response.json().get("message", "Request failed")
            except Exception:
                message = "Request failed"
            raise Exception(f"HTTP {response.status_code}: {message}") from err
        except Exception as err:
            raise Exception(
                "Network error or request failed without a response."
            ) from err

    def get(self, url: str, req_options: Optional[HTTPRequestOptions] = None):
        """
        Make a GET request.
        """
        if req_options is None:
            req_options = HTTPRequestOptions(send_through_proxy=True)
        return self.request("GET", url, None, req_options)

    def post(
        self, url: str, data: Any, req_options: Optional[HTTPRequestOptions] = None
    ):
        """
        Make a POST request.
        """
        return self.request("POST", url, data, req_options)

    def put(
        self, url: str, data: Any, req_options: Optional[HTTPRequestOptions] = None
    ):
        """
        Make a PUT request.
        """
        return self.request("PUT", url, data, req_options)

    def delete(
        self, url: str, data: Any, req_options: Optional[HTTPRequestOptions] = None
    ):
        """
        Make a DELETE request.
        """
        return self.request("DELETE", url, data, req_options)



"""PaymentsRequestHandler adds payments validation & credit burning on top of DefaultRequestHandler."""

from __future__ import annotations

import asyncio
import base64
import logging
from typing import Any, Dict

import httpx
from a2a.server.events.event_consumer import EventConsumer
from a2a.server.events.event_queue import EventQueue
from a2a.server.request_handlers.default_request_handler import DefaultRequestHandler
from a2a.server.tasks.result_aggregator import ResultAggregator
from a2a.types import (
    Message,
    MessageSendParams,
    Task,
    TaskIdParams,
    TaskStatusUpdateEvent,
)
from payments_py.common.payments_error import PaymentsError
from payments_py.payments import Payments

from .types import HttpRequestContext


_TERMINAL_STATES = {
    "completed",
    "failed",
    "canceled",
    "rejected",
}


class PaymentsRequestHandler(DefaultRequestHandler):  # noqa: D101
    """Extend DefaultRequestHandler adding credit validation & burning."""

    # ------------------------------------------------------------------
    # Lifecycle --------------------------------------------------------
    # ------------------------------------------------------------------
    def __init__(
        self,
        *,
        agent_card: Any,  # a2a.types.AgentCard
        task_store: Any,
        agent_executor: Any,
        payments_service: Payments,
        queue_manager: Any | None = None,
        push_config_store: Any | None = None,
        push_sender: Any | None = None,
        request_context_builder: Any | None = None,
        async_execution: bool = False,
    ) -> None:
        super().__init__(
            agent_executor=agent_executor,
            task_store=task_store,
            queue_manager=queue_manager,
            push_config_store=push_config_store,
            push_sender=push_sender,
            request_context_builder=request_context_builder,
        )
        self._agent_card = agent_card
        self._payments = payments_service
        self._async_execution = async_execution
        self._http_ctx_by_task: Dict[str, HttpRequestContext] = {}
        self._http_ctx_by_message: Dict[str, HttpRequestContext] = {}

    # ------------------------------------------------------------------
    # Context helpers (called by middleware) ---------------------------
    # ------------------------------------------------------------------
    def set_http_ctx_for_task(
        self, task_id: str, ctx: HttpRequestContext
    ) -> None:  # noqa: D401
        self._http_ctx_by_task[task_id] = ctx

    def set_http_ctx_for_message(
        self, message_id: str, ctx: HttpRequestContext
    ) -> None:  # noqa: D401
        self._http_ctx_by_message[message_id] = ctx

    # ------------------------------------------------------------------
    def _get_http_ctx(
        self, task_id: str | None, message_id: str | None
    ) -> HttpRequestContext | None:  # noqa: D401
        if task_id and task_id in self._http_ctx_by_task:
            return self._http_ctx_by_task[task_id]
        if message_id and message_id in self._http_ctx_by_message:
            return self._http_ctx_by_message[message_id]
        return None

    def _migrate_http_ctx_from_message_to_task(
        self, message_id: str, task_id: str
    ) -> None:  # noqa: D401
        """Migrate HTTP context from messageId to taskId when task is created."""
        if message_id in self._http_ctx_by_message:
            ctx = self._http_ctx_by_message.pop(message_id)
            self._http_ctx_by_task[task_id] = ctx

    def delete_http_ctx_for_task(self, task_id: str) -> None:  # noqa: D401
        self._http_ctx_by_task.pop(task_id, None)

    def delete_http_ctx_for_message(self, message_id: str) -> None:  # noqa: D401
        self._http_ctx_by_message.pop(message_id, None)

    def migrate_http_ctx_from_message_to_task(
        self, message_id: str, task_id: str
    ) -> None:  # noqa: D401
        self._migrate_http_ctx_from_message_to_task(message_id, task_id)

    async def validate_request(
        self,
        agent_id: str,
        bearer_token: str,
        url_requested: str,
        http_method_requested: str,
    ) -> Any:
        """
        Validates a request using the payments service.
        This method is used by the middleware to validate credits before processing requests.

        Args:
            agent_id: The agent ID to validate
            bearer_token: The bearer token for authentication
            url_requested: The URL being requested
            http_method_requested: The HTTP method being used

        Returns:
            The validation result containing agentRequestId and other validation data

        Raises:
            PaymentsError: If validation fails
        """
        # Use run_in_executor since start_processing_request is synchronous
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(
            None,
            lambda: self._payments.requests.start_processing_request(  # type: ignore[attr-defined]
                agent_id,
                bearer_token,
                url_requested,
                http_method_requested,
            ),
        )

    # ------------------------------------------------------------------
    # Overrides --------------------------------------------------------
    # ------------------------------------------------------------------
    async def on_message_send(
        self,
        params: MessageSendParams,
        context: Any | None = None,
    ) -> Task | Message:  # noqa: D401
        """Override sendMessage to add payments validation & credit burning like TypeScript."""
        # Validate required parameters
        if not params.message:
            raise PaymentsError.bad_request("message is required.")
        if not params.message.message_id:
            raise PaymentsError.bad_request("message.messageId is required.")

        # Get HTTP context for the task or message
        prev_task_id = params.message.task_id
        http_ctx = None
        if prev_task_id:
            http_ctx = self._get_http_ctx(prev_task_id, None)
        else:
            http_ctx = self._get_http_ctx(None, params.message.message_id)

        if http_ctx is None:
            raise PaymentsError.unauthorized(
                "HTTP context missing for request; bearer token not found."
            )

        # Get agentId from agent card (like TypeScript)
        agent_card = await self.get_agent_card()

        # Extract agentId from payment extension (handle both dict and SimpleNamespace)
        agent_id = None
        if hasattr(agent_card, "get"):
            # Dictionary-style access
            extensions = agent_card.get("capabilities", {}).get("extensions", [])
        else:
            # Object-style access (SimpleNamespace)
            capabilities = getattr(agent_card, "capabilities", None)
            extensions = getattr(capabilities, "extensions", []) if capabilities else []

        for ext in extensions:
            if (hasattr(ext, "get") and ext.get("uri") == "urn:nevermined:payment") or (
                hasattr(ext, "uri") and ext.uri == "urn:nevermined:payment"
            ):
                # Handle both dict and SimpleNamespace for params
                if hasattr(ext, "get"):
                    agent_id = ext.get("params", {}).get("agentId")
                else:
                    ext_params = getattr(ext, "params", None)
                    agent_id = (
                        getattr(ext_params, "agentId", None) if ext_params else None
                    )
                break

        if not agent_id:
            raise PaymentsError.internal("Agent ID not found in payment extension.")

        # Setup message execution (equivalent to TS setup)
        (
            task_manager,
            task_id,
            queue,
            result_aggregator,
            producer_task,
        ) = await self._setup_message_execution(params, context)

        # migrate HTTP context from message to task
        if not prev_task_id:
            self._migrate_http_ctx_from_message_to_task(
                params.message.message_id, task_id
            )
            # Update the message with the new taskId
            params.message.task_id = task_id

        consumer = EventConsumer(queue)
        producer_task.add_done_callback(consumer.agent_task_callback)

        # Determine if execution should be blocking
        blocking = True
        if (
            hasattr(params, "configuration")
            and params.configuration
            and params.configuration.blocking is False
        ):
            blocking = False

        interrupted_or_non_blocking = False
        try:
            # Both blocking and non-blocking use the same method, but with different
            # early return behavior
            (result, interrupted_or_non_blocking) = (
                await self._consume_and_burn_credits(
                    result_aggregator, consumer, http_ctx, blocking
                )
            )

            if not result:
                raise PaymentsError.internal(
                    "Agent execution finished without a result, and no task context found."
                )

            if isinstance(result, Task):
                self._validate_task_id_match(task_id, result.id)

            await self._send_push_notification_if_needed(task_id, result_aggregator)

            return result

        except Exception as e:
            logging.getLogger(__name__).error(f"Agent execution failed. Error: {e}")
            raise
        finally:
            # Cleanup like parent implementation
            if interrupted_or_non_blocking:
                # For non-blocking mode, schedule background cleanup (like parent SDK
                # does)
                asyncio.create_task(self._cleanup_producer(producer_task, task_id))
            else:
                await self._cleanup_producer(producer_task, task_id)

    async def _monitor_for_final_events(
        self,
        task_id: str,
        http_ctx: HttpRequestContext,
        producer_task: "asyncio.Task[None]",
        queue: "EventQueue",
    ) -> None:
        """Monitor for final events with credit burning in non-blocking mode."""
        try:
            print(f"[DEBUG] Starting monitor for final events for {task_id}")

            # Wait for the producer task to complete (this publishes the final event)
            await producer_task
            print(f"[DEBUG] Producer task completed for {task_id}")

            # Create a new consumer to check for any remaining events in the queue
            final_consumer = EventConsumer(queue)

            print(
                f"[DEBUG] Created final consumer for {task_id}, checking for final events"
            )

            # Check for any remaining events (with timeout)
            try:
                async for event in final_consumer.consume_all():
                    print(
                        f"[DEBUG] monitor_for_final_events got event: {type(event)} - {getattr(event, 'kind', 'no-kind')} - final: {getattr(event, 'final', 'no-final')}"
                    )

                    if (
                        isinstance(event, TaskStatusUpdateEvent)
                        and event.final is True
                        and hasattr(event, "metadata")
                        and event.metadata
                        and event.metadata.get("creditsUsed") is not None
                        and http_ctx.bearer_token
                    ):
                        print(
                            f"[DEBUG] Found final event with creditsUsed: {event.metadata.get('creditsUsed')}"
                        )
                        await self._handle_task_finalization_from_event(event, http_ctx)
                        break

            except Exception as e:
                print(f"[DEBUG] Error consuming final events: {e}")

            # Give a bit more time for any delayed events
            await asyncio.sleep(0.2)

        except asyncio.CancelledError:
            print(f"[DEBUG] Monitor task cancelled for {task_id}")
        except Exception as e:
            logging.getLogger(__name__).warning(
                f"Monitor task failed for {task_id}: {e}"
            )
            print(f"[DEBUG] Monitor task exception: {e}")
        finally:
            # Clean up HTTP context
            self.delete_http_ctx_for_task(task_id)
            print(f"[DEBUG] Monitor cleanup completed for {task_id}")

    async def _delayed_cleanup_for_non_blocking(
        self,
        producer_task: "asyncio.Task[None]",
        task_id: str,
        http_ctx: HttpRequestContext,
        result_aggregator: ResultAggregator,
    ) -> None:
        """Wait for task completion in non-blocking mode, then cleanup."""
        try:
            # Wait for the producer task to complete naturally (don't cancel it)
            await producer_task

            # After producer completes, ensure we've processed all events
            # Give some time for final events to be processed
            await asyncio.sleep(0.1)

        except asyncio.CancelledError:
            # If cancelled, that's expected behavior
            pass
        except Exception as e:
            logging.getLogger(__name__).warning(
                f"Producer task failed for {task_id}: {e}"
            )
        finally:
            # Clean up HTTP context
            self.delete_http_ctx_for_task(task_id)

    async def _cleanup_producer(self, producer_task, task_id: str) -> None:
        """Cleanup producer task (from parent implementation)."""
        if not producer_task.done():
            producer_task.cancel()
            try:
                await producer_task
            except asyncio.CancelledError:
                pass

    async def _consume_and_burn_credits(
        self,
        result_aggregator: ResultAggregator,
        consumer: EventConsumer,
        http_ctx: HttpRequestContext,
        blocking: bool = True,
    ) -> tuple[Task | Message, bool]:
        """Process events with credit burning (like TypeScript processEventsWithFinalization).

        This mirrors TypeScript processEventsWithFinalization with:
        - Credit burning on final events with creditsUsed metadata
        - Background processing continuation when interrupted (via SDK's _continue_consuming)
        """
        # Store the original consume_all method before replacing it
        original_consume_all = consumer.consume_all

        # Create a custom event processor that intercepts events for credit burning
        async def credit_burning_event_processor():
            async for event in original_consume_all():
                print(
                    f"[DEBUG] credit_burning_event_processor got event: {type(event)} - {getattr(event, 'kind', 'no-kind')} - final: {getattr(event, 'final', 'no-final')} - metadata: {getattr(event, 'metadata', 'no-metadata')}"
                )

                # Handle credit burning on TaskStatusUpdateEvent (like TypeScript
                # handleTaskFinalization)
                if (
                    isinstance(event, TaskStatusUpdateEvent)
                    and event.final is True
                    and hasattr(event, "metadata")
                    and event.metadata
                    and event.metadata.get("creditsUsed") is not None
                    and http_ctx.bearer_token
                ):
                    print(
                        f"[DEBUG] Processing credit burning for final event: {event.metadata.get('creditsUsed')} credits"
                    )
                    await self._handle_task_finalization_from_event(event, http_ctx)

                yield event

        # Replace consumer's consume_all with our credit burning processor
        consumer.consume_all = credit_burning_event_processor

        try:
            if blocking:
                # Blocking mode: intercept events in main flow (current working
                # approach)
                return await result_aggregator.consume_and_break_on_interrupt(
                    consumer, blocking=blocking
                )
            else:
                # Non-blocking mode: don't intercept main flow, but intercept background processing
                # First, restore original consumer to avoid interfering with SDK's early
                # return logic
                consumer.consume_all = original_consume_all

                # Intercept the _continue_consuming method for background credit burning
                original_continue_consuming = result_aggregator._continue_consuming

                async def background_credit_burning_processor(event_stream):
                    """Process background events with credit burning."""
                    async for event in event_stream:
                        print(
                            f"[DEBUG] background_credit_burning_processor got event: {type(event)} - {getattr(event, 'kind', 'no-kind')} - final: {getattr(event, 'final', 'no-final')}"
                        )

                        # Process the event normally (like original _continue_consuming)
                        await result_aggregator.task_manager.process(event)

                        # Check for credit burning on final events
                        if (
                            isinstance(event, TaskStatusUpdateEvent)
                            and event.final is True
                            and hasattr(event, "metadata")
                            and event.metadata
                            and event.metadata.get("creditsUsed") is not None
                            and http_ctx.bearer_token
                        ):
                            print(
                                f"[DEBUG] Background credit burning for: {event.metadata.get('creditsUsed')} credits"
                            )
                            await self._handle_task_finalization_from_event(
                                event, http_ctx
                            )

                # Replace _continue_consuming temporarily
                result_aggregator._continue_consuming = (
                    background_credit_burning_processor
                )

                try:
                    # Let SDK do its normal non-blocking flow
                    result = await result_aggregator.consume_and_break_on_interrupt(
                        consumer, blocking=blocking
                    )
                    return result
                finally:
                    # Restore original _continue_consuming
                    result_aggregator._continue_consuming = original_continue_consuming
        finally:
            # Restore original consume_all if it wasn't already restored
            if consumer.consume_all != original_consume_all:
                consumer.consume_all = original_consume_all

    async def _handle_task_finalization_from_event(
        self, event: TaskStatusUpdateEvent, http_ctx: HttpRequestContext
    ) -> None:
        """Handle credit burning from TaskStatusUpdateEvent (like TypeScript handleTaskFinalization)."""
        if not event.metadata or not event.metadata.get("creditsUsed"):
            return

        print(f"[DEBUG] Handling task finalization from event: {event}")

        credits_used = event.metadata["creditsUsed"]
        try:
            loop = asyncio.get_running_loop()
            await loop.run_in_executor(
                None,
                lambda: self._payments.requests.redeem_credits_from_request(  # type: ignore[attr-defined]
                    http_ctx.validation["agentRequestId"],
                    http_ctx.bearer_token,
                    int(credits_used),
                ),
            )
        except Exception:  # noqa: BLE001
            # Swallow redeem errors (non-blocking)
            pass

    # ------------------------------------------------------------------
    # Streaming override ------------------------------------------------
    # ------------------------------------------------------------------
    async def on_message_send_stream(
        self,
        params: MessageSendParams,
        context: Any | None = None,
    ) -> Any:  # noqa: D401
        """Override streaming to handle credit burning when TaskStatusUpdateEvent arrives."""
        # Retrieve HTTP context
        http_ctx = self._get_http_ctx(
            params.message.task_id if params.message else None,
            params.message.message_id if params.message else None,
        )
        if http_ctx is None:
            raise PaymentsError.unauthorized(
                "HTTP context missing for request; bearer token not found."
            )

        # Call parent streaming method and process events
        # type: ignore[arg-type]
        async for event in super().on_message_send_stream(params, context):
            # Handle credit burning on final status updates
            if (
                isinstance(event, dict)
                and event.get("kind") == "status-update"
                and event.get("final") is True
                and event.get("metadata", {}).get("creditsUsed") is not None
                and http_ctx.bearer_token
            ):
                credits_used = event["metadata"]["creditsUsed"]
                try:
                    loop = asyncio.get_running_loop()
                    await loop.run_in_executor(
                        None,
                        lambda: self._payments.requests.redeem_credits_from_request(  # type: ignore[attr-defined]
                            http_ctx.validation["agentRequestId"],
                            http_ctx.bearer_token,
                            int(credits_used),
                        ),
                    )
                except Exception:  # noqa: BLE001
                    # Swallow redeem errors (non-blocking)
                    pass

            # Handle push notifications on final status updates
            if (
                isinstance(event, dict)
                and event.get("kind") == "status-update"
                and event.get("final") is True
                and event.get("status", {}).get("state") in _TERMINAL_STATES
            ):
                try:
                    task_id = event.get("taskId")
                    state = event["status"]["state"]
                    push_cfg = await self.on_get_task_push_notification_config(
                        TaskIdParams(id=task_id)
                    )
                    if push_cfg:
                        await self._send_push_notification(
                            task_id,
                            state,
                            push_cfg["pushNotificationConfig"],
                        )
                except Exception:  # noqa: BLE001
                    # Swallow push notification errors (non-blocking)
                    pass

            yield event

    async def _send_push_notification_if_needed(
        self, task_id: str, result_aggregator: Any
    ) -> None:
        """Send push notification if needed for completed task."""
        try:
            # Get the final result to check if task is completed
            task = await result_aggregator.task_manager.get_task()
            if task and task.status.state in _TERMINAL_STATES:
                # Get push notification config
                push_cfg = await self.on_get_task_push_notification_config(
                    TaskIdParams(id=task_id)
                )
                if push_cfg and "pushNotificationConfig" in push_cfg:
                    await self._send_push_notification(
                        task_id,
                        task.status.state,
                        push_cfg["pushNotificationConfig"],
                    )
        except Exception:  # noqa: BLE001
            # Swallow push notification errors (non-blocking)
            pass

    # ------------------------------------------------------------------
    # Push notification helper -----------------------------------------
    # ------------------------------------------------------------------
    async def _send_push_notification(
        self,
        task_id: str,
        state: str,
        push_notification_config: Dict[str, Any],
        payload: Dict[str, Any] | None = None,
    ) -> None:
        """Send HTTP push notification (best-effort)."""
        headers: Dict[str, str] = {"Content-Type": "application/json"}
        if auth := push_notification_config.get("authentication"):
            schemes = auth.get("schemes", [])
            creds = auth.get("credentials")
            if "basic" in schemes:
                headers["Authorization"] = (
                    "Basic " + base64.b64encode(creds.encode()).decode()
                )
            elif "bearer" in schemes:
                headers["Authorization"] = f"Bearer {creds}"
            elif "custom" in schemes and isinstance(creds, dict):
                headers.update(creds)

        data = {
            "taskId": task_id,
            "state": state,
            "payload": payload or {},
        }

        try:
            async with httpx.AsyncClient() as client:
                await client.post(
                    push_notification_config["url"],
                    json=data,
                    headers=headers,
                    timeout=5.0,
                )
        except Exception:  # noqa: BLE001
            pass  # ignore push errors

    # ------------------------------------------------------------------
    # Agent card accessor -----------------------------------------------
    # ------------------------------------------------------------------
    async def get_agent_card(self) -> Any:  # noqa: D401
        return self._agent_card



"""
Utility functions for creating and managing payment plans.
"""

from payments_py.common.types import (
    PlanCreditsConfig,
    PlanCreditsType,
    PlanPriceConfig,
    PlanPriceType,
    PlanRedemptionType,
    Address,
)
from payments_py.environments import ZeroAddress
from payments_py.utils import is_ethereum_address

# Duration constants in seconds
ONE_DAY_DURATION = 86_400  # 24 * 60 * 60 seconds
ONE_WEEK_DURATION = 604_800  # 7 * 24 * 60 * 60 seconds
ONE_MONTH_DURATION = (
    2_629_746  # (365.25 days/year ÷ 12 months/year) × 24 × 60 × 60 ≈ 2,629,746 seconds
)
ONE_YEAR_DURATION = 31_557_600  # 365.25 * 24 * 60 * 60 seconds


def get_fiat_price_config(amount: int, receiver: Address) -> PlanPriceConfig:
    """
    Get a fixed fiat price configuration for a plan.

    Args:
        amount: The amount in the smallest unit of the fiat currency
        receiver: The address that will receive the payment

    Returns:
        A PlanPriceConfig object configured for fiat payments

    Raises:
        ValueError: If the receiver address is not a valid Ethereum address
    """
    if not is_ethereum_address(receiver):
        raise ValueError(f"Receiver address {receiver} is not a valid Ethereum address")
    return PlanPriceConfig(
        price_type=PlanPriceType.FIXED_FIAT_PRICE,
        token_address=ZeroAddress,
        amounts=[amount],
        receivers=[receiver],
        contract_address=ZeroAddress,
        fee_controller=ZeroAddress,
    )


def get_crypto_price_config(
    amount: int, receiver: Address, token_address: Address = ZeroAddress
) -> PlanPriceConfig:
    """
    Get a fixed crypto price configuration for a plan.

    Args:
        amount: The amount in the smallest unit of the token
        receiver: The address that will receive the payment
        token_address: The address of the token to use for payment (defaults to native token)

    Returns:
        A PlanPriceConfig object configured for crypto payments

    Raises:
        ValueError: If the receiver address is not a valid Ethereum address
    """
    if not is_ethereum_address(receiver):
        raise ValueError(f"Receiver address {receiver} is not a valid Ethereum address")
    return PlanPriceConfig(
        price_type=PlanPriceType.FIXED_PRICE,
        token_address=token_address,
        amounts=[amount],
        receivers=[receiver],
        contract_address=ZeroAddress,
        fee_controller=ZeroAddress,
    )


def get_erc20_price_config(
    amount: int, token_address: Address, receiver: Address
) -> PlanPriceConfig:
    """
    Get a fixed ERC20 token price configuration for a plan.

    Args:
        amount: The amount in the smallest unit of the ERC20 token
        token_address: The address of the ERC20 token
        receiver: The address that will receive the payment

    Returns:
        A PlanPriceConfig object configured for ERC20 token payments
    """
    return get_crypto_price_config(amount, receiver, token_address)


def get_free_price_config() -> PlanPriceConfig:
    """
    Get a free price configuration for a plan.

    Returns:
        A PlanPriceConfig object configured for free plans
    """
    return PlanPriceConfig(
        price_type=PlanPriceType.FIXED_PRICE,
        token_address=ZeroAddress,
        amounts=[],
        receivers=[],
        contract_address=ZeroAddress,
        fee_controller=ZeroAddress,
    )


def get_native_token_price_config(amount: int, receiver: Address) -> PlanPriceConfig:
    """
    Get a fixed native token price configuration for a plan.

    Args:
        amount: The amount in the smallest unit of the native token
        receiver: The address that will receive the payment

    Returns:
        A PlanPriceConfig object configured for native token payments
    """
    return get_crypto_price_config(amount, receiver, ZeroAddress)


def get_expirable_duration_config(duration_of_plan: int) -> PlanCreditsConfig:
    """
    Get an expirable duration configuration for a plan.

    Args:
        duration_of_plan: The duration of the plan in seconds

    Returns:
        A PlanCreditsConfig object configured for expirable duration
    """
    return PlanCreditsConfig(
        credits_type=PlanCreditsType.EXPIRABLE,
        redemption_type=PlanRedemptionType.ONLY_OWNER,
        proof_required=False,
        duration_secs=duration_of_plan,
        amount=1,
        min_amount=1,
        max_amount=1,
    )


def get_non_expirable_duration_config() -> PlanCreditsConfig:
    """
    Get a non-expirable duration configuration for a plan.

    Returns:
        A PlanCreditsConfig object configured for non-expirable duration
    """
    return get_expirable_duration_config(0)


def get_fixed_credits_config(
    credits_granted: int, credits_per_request: int = 1
) -> PlanCreditsConfig:
    """
    Get a fixed credits configuration for a plan.

    Args:
        credits_granted: The total number of credits granted
        credits_per_request: The number of credits consumed per request (default: 1)

    Returns:
        A PlanCreditsConfig object configured for fixed credits
    """
    return PlanCreditsConfig(
        credits_type=PlanCreditsType.FIXED,
        redemption_type=PlanRedemptionType.ONLY_OWNER,
        proof_required=False,
        duration_secs=0,
        amount=credits_granted,
        min_amount=credits_per_request,
        max_amount=credits_per_request,
    )


def get_dynamic_credits_config(
    credits_granted: int,
    min_credits_per_request: int = 1,
    max_credits_per_request: int = 1,
) -> PlanCreditsConfig:
    """
    Get a dynamic credits configuration for a plan.

    Args:
        credits_granted: The total number of credits granted
        min_credits_per_request: The minimum number of credits consumed per request (default: 1)
        max_credits_per_request: The maximum number of credits consumed per request (default: 1)

    Returns:
        A PlanCreditsConfig object configured for dynamic credits
    """
    return PlanCreditsConfig(
        credits_type=PlanCreditsType.DYNAMIC,
        redemption_type=PlanRedemptionType.ONLY_OWNER,
        proof_required=False,
        duration_secs=0,
        amount=credits_granted,
        min_amount=min_credits_per_request,
        max_amount=max_credits_per_request,
    )


def set_redemption_type(
    credits_config: PlanCreditsConfig, redemption_type: PlanRedemptionType
) -> PlanCreditsConfig:
    """
    Set the redemption type for a credits configuration.

    Args:
        credits_config: The credits configuration to modify
        redemption_type: The new redemption type

    Returns:
        A new PlanCreditsConfig with the updated redemption type
    """
    return PlanCreditsConfig(
        credits_type=credits_config.credits_type,
        redemption_type=redemption_type,
        proof_required=credits_config.proof_required,
        duration_secs=credits_config.duration_secs,
        amount=credits_config.amount,
        min_amount=credits_config.min_amount,
        max_amount=credits_config.max_amount,
    )


def set_proof_required(
    credits_config: PlanCreditsConfig, proof_required: bool = True
) -> PlanCreditsConfig:
    """
    Set whether proof is required for a credits configuration.

    Args:
        credits_config: The credits configuration to modify
        proof_required: Whether proof is required (default: True)

    Returns:
        A new PlanCreditsConfig with the updated proof requirement
    """
    return PlanCreditsConfig(
        credits_type=credits_config.credits_type,
        redemption_type=credits_config.redemption_type,
        proof_required=proof_required,
        duration_secs=credits_config.duration_secs,
        amount=credits_config.amount,
        min_amount=credits_config.min_amount,
        max_amount=credits_config.max_amount,
    )



"""
Types for MCP paywall functionality.
"""

from typing import Any, Callable, Dict, TypedDict, Union


class AuthResult(TypedDict):
    """Result returned by authentication routines."""

    requestId: str
    token: str
    agentId: str
    logicalUrl: str


CreditsOption = Union[int, Callable[[Dict[str, Any]], int]]


class BasePaywallOptions(TypedDict, total=False):
    """Common paywall options shared by all handler kinds."""

    name: str
    credits: CreditsOption
    onRedeemError: str  # 'ignore' | 'propagate'


class ToolOptions(BasePaywallOptions):
    """Paywall options for a tool handler."""

    kind: str  # 'tool'


class ResourceOptions(BasePaywallOptions):
    """Paywall options for a resource handler."""

    kind: str  # 'resource'


class PromptOptions(BasePaywallOptions):
    """Paywall options for a prompt handler."""

    kind: str  # 'prompt'


PaywallOptions = Union[ToolOptions, ResourceOptions, PromptOptions]



"""
MCP integration for Nevermined Payments (Python).

Exposes the factory ``build_mcp_integration`` and small helpers.
"""

from .index import build_mcp_integration, MCPIntegration  # noqa: F401
from .utils.extra import (  # noqa: F401
    build_extra_from_http_headers,
    build_extra_from_http_request,
    build_extra_from_fastmcp_context,
)

__all__ = [
    "build_mcp_integration",
    "build_extra_from_http_headers",
    "build_extra_from_http_request",
    "build_extra_from_fastmcp_context",
    "MCPIntegration",
]



"""
Helpers to build logical URLs for MCP tools, resources and prompts.
"""

from typing import Any, Dict
from urllib.parse import urlencode


def build_logical_tool_url(server_name: str, tool_name: str, args: Any) -> str:
    """Build a logical MCP URL for a tool call.

    Args:
        server_name: Logical MCP server name.
        tool_name: Tool identifier.
        args: Arguments dictionary to be encoded in the query string.

    Returns:
        Logical URL string.
    """
    query = ""
    try:
        obj = args if isinstance(args, dict) else {}
        params = {}
        for k, v in obj.items():
            params[str(k)] = v if isinstance(v, str) else json_dumps_safe(v)
        s = urlencode(params)
        query = f"?{s}" if s else ""
    except Exception:
        pass
    return f"mcp://{server_name}/tools/{tool_name}{query}"


def build_logical_resource_url(
    server_name: str, resource_name: str, variables: Dict[str, Any]
) -> str:
    """Build a logical MCP URL for a resource retrieval.

    Args:
        server_name: Logical MCP server name.
        resource_name: Resource identifier.
        variables: Variables dictionary to be encoded in the query string.

    Returns:
        Logical URL string.
    """
    query = ""
    try:
        params = {}
        for k, v in (variables or {}).items():
            params[str(k)] = v[0] if isinstance(v, list) and v else str(v)
        s = urlencode(params)
        query = f"?{s}" if s else ""
    except Exception:
        pass
    return f"mcp://{server_name}/resources/{resource_name}{query}"


def build_logical_meta_url(server_name: str, method: str) -> str:
    """Build a logical MCP URL for a meta endpoint.

    Args:
        server_name: Logical MCP server name.
        method: Meta operation name.

    Returns:
        Logical URL string.
    """
    safe = method if isinstance(method, str) else "unknown"
    return f"mcp://{server_name}/meta/{safe}"


def build_logical_url(options: Dict[str, Any]) -> str:
    """Build a logical MCP URL from generic options.

    Args:
        options: Dictionary with keys kind, serverName, name and argsOrVars.

    Returns:
        Logical URL string.
    """
    kind = options.get("kind")
    server_name = options.get("serverName")
    name = options.get("name")
    args_or_vars = options.get("argsOrVars")
    if kind == "resource":
        return build_logical_resource_url(server_name, name, args_or_vars or {})
    return build_logical_tool_url(server_name, name, args_or_vars)


def json_dumps_safe(value: Any) -> str:
    """Serialize a Python value to JSON without raising exceptions.

    Falls back to ``str(value)`` when standard JSON serialization fails.

    Args:
        value: Any Python value.

    Returns:
        JSON string or stringified value.
    """
    try:
        import json

        return json.dumps(value, separators=(",", ":"))
    except Exception:
        return str(value)



"""
Type definitions for the Nevermined Payments protocol.
"""

from typing import Optional, List, Dict, Any
from pydantic import BaseModel, Field
from enum import Enum

# Address type alias
Address = str


class PaymentOptions(BaseModel):
    """
    Options for initializing the Payments class.
    """

    environment: str
    nvm_api_key: Optional[str] = None
    return_url: Optional[str] = None
    app_id: Optional[str] = None
    version: Optional[str] = None
    headers: Optional[Dict[str, str]] = None


class Endpoint(BaseModel):
    """
    Endpoint for a service. Dict with HTTP verb as key and URL as value.
    """

    verb: str
    url: str


class AuthType(str, Enum):
    """
    Allowed authentication types for AgentAPIAttributes.
    """

    NONE = "none"
    BASIC = "basic"
    OAUTH = "oauth"
    BEARER = "bearer"


class AgentAPIAttributes(BaseModel):
    """
    API attributes for an agent.
    """

    endpoints: List[Endpoint]
    open_endpoints: Optional[List[str]] = None
    open_api_url: Optional[str] = None
    auth_type: Optional[AuthType] = AuthType.NONE
    username: Optional[str] = None
    password: Optional[str] = None
    token: Optional[str] = None
    api_key: Optional[str] = None
    headers: Optional[Dict[str, str]] = None


class AgentMetadata(BaseModel):
    """
    Metadata for an agent.
    """

    name: str
    description: Optional[str] = None
    author: Optional[str] = None
    license: Optional[str] = None
    tags: Optional[List[str]] = None
    integration: Optional[str] = None
    sample_link: Optional[str] = None
    api_description: Optional[str] = None
    date_created: Optional[str] = None


class PlanMetadata(AgentMetadata):
    """
    Metadata for a payment plan, extends AgentMetadata.
    """

    is_trial_plan: Optional[bool] = False


class PlanPriceType(Enum):
    """
    Different types of prices that can be configured for a plan.
    0 - FIXED_PRICE, 1 - FIXED_FIAT_PRICE, 2 - SMART_CONTRACT_PRICE
    """

    FIXED_PRICE = 0
    FIXED_FIAT_PRICE = 1
    SMART_CONTRACT_PRICE = 2


class PlanCreditsType(Enum):
    """
    Different types of credits that can be obtained when purchasing a plan.
    0 - EXPIRABLE, 1 - FIXED, 2 - DYNAMIC
    """

    EXPIRABLE = 0
    FIXED = 1
    DYNAMIC = 2


class PlanRedemptionType(Enum):
    """
    Different types of redemptions criterias that can be used when redeeming credits.
    0 - ONLY_GLOBAL_ROLE, 1 - ONLY_OWNER, 2 - ONLY_PLAN_ROLE
    """

    ONLY_GLOBAL_ROLE = 0
    ONLY_OWNER = 1
    ONLY_PLAN_ROLE = 2


class PlanPriceConfig(BaseModel):
    """
    Definition of the price configuration for a Payment Plan.
    """

    price_type: PlanPriceType
    token_address: Optional[str] = None
    amounts: List[int] = Field(default_factory=list)
    receivers: List[str] = Field(default_factory=list)
    contract_address: Optional[str] = None
    fee_controller: Optional[str] = None


class PlanCreditsConfig(BaseModel):
    """
    Definition of the credits configuration for a payment plan.
    """

    credits_type: PlanCreditsType
    redemption_type: PlanRedemptionType
    proof_required: bool
    duration_secs: int
    amount: int
    min_amount: int
    max_amount: int
    nft_address: Optional[str] = None


class PlanBalance(BaseModel):
    """
    Balance information for a payment plan.
    """

    plan_id: str
    holder_address: str
    balance: int
    credits_contract: str
    is_subscriber: bool


class PaginationOptions(BaseModel):
    """
    Options for pagination in API requests to the Nevermined API.
    """

    sort_by: Optional[str] = None
    sort_order: str = "desc"
    page: int = 1
    offset: int = 10


class AgentTaskStatus(str, Enum):
    """
    Status of an agent task.
    """

    SUCCESS = "SUCCESS"
    FAILED = "FAILED"
    PENDING = "PENDING"


class TrackAgentSubTaskDto(BaseModel):
    """
    Data transfer object for tracking agent sub tasks.
    """

    agent_request_id: str
    credits_to_redeem: Optional[int] = 0
    tag: Optional[str] = None
    description: Optional[str] = None
    status: Optional[AgentTaskStatus] = None


class StartAgentRequest(BaseModel):
    """
    Information about the initialization of an agent request.
    """

    agent_request_id: str
    balance: PlanBalance
    url_matching: str
    verb_matching: str


class AgentAccessCredentials(BaseModel):
    """
    Access credentials for an agent.
    """

    access_token: str
    proxies: Optional[List[str]] = None


class NvmAPIResult(BaseModel):
    """
    Result of a Nevermined API operation.
    """

    success: bool
    message: Optional[str] = None
    tx_hash: Optional[str] = None
    http_status: Optional[int] = None
    data: Optional[Dict[str, Any]] = None
    when: Optional[str] = None



"""
The PlansAPI class provides methods to register and interact with payment plans on Nevermined.
"""

import requests
from typing import Dict, Any, Optional
from payments_py.common.payments_error import PaymentsError
from payments_py.common.types import (
    PaymentOptions,
    PlanMetadata,
    PlanPriceConfig,
    PlanCreditsConfig,
    PlanCreditsType,
    PlanRedemptionType,
    PaginationOptions,
)
from payments_py.api.base_payments import BasePaymentsAPI
from payments_py.api.nvm_api import (
    API_URL_REGISTER_PLAN,
    API_URL_GET_PLAN,
    API_URL_PLAN_BALANCE,
    API_URL_ORDER_PLAN,
    API_URL_MINT_PLAN,
    API_URL_MINT_EXPIRABLE_PLAN,
    API_URL_BURN_PLAN,
    API_URL_GET_PLAN_AGENTS,
    API_URL_STRIPE_CHECKOUT,
    API_URL_REDEEM_PLAN,
)
from payments_py.utils import get_random_big_int, is_ethereum_address
from payments_py import plans as plan_utils


class PlansAPI(BasePaymentsAPI):
    """
    The PlansAPI class provides methods to register and interact with payment plans on Nevermined.
    """

    @classmethod
    def get_instance(cls, options: PaymentOptions) -> "PlansAPI":
        """
        Get a singleton instance of the PlansAPI class.

        Args:
            options: The options to initialize the payments class

        Returns:
            The instance of the PlansAPI class
        """
        return cls(options)

    def register_plan(
        self,
        plan_metadata: PlanMetadata,
        price_config: PlanPriceConfig,
        credits_config: PlanCreditsConfig,
        nonce: Optional[int] = None,
    ) -> Dict[str, str]:
        """
        Allows an AI Builder to create a Payment Plan on Nevermined in a flexible manner.
        A Nevermined Credits Plan limits access based on plan usage.
        With them, AI Builders control the number of requests that can be made to an agent or service.
        Every time a user accesses any resource associated with the Payment Plan, the usage consumes from a capped amount of credits.
        When the user consumes all the credits, the plan automatically expires and the user needs to top up to continue using the service.

        Args:
            plan_metadata: Plan metadata
            price_config: Plan price configuration
            credits_config: Plan credits configuration
            nonce: Optional nonce for the transaction

        Returns:
            The unique identifier of the plan (Plan ID) of the newly created plan

        Raises:
            PaymentsError: If registration fails
        """
        if nonce is None:
            nonce = get_random_big_int()

        body = {
            "metadataAttributes": self.pydantic_to_dict(plan_metadata),
            "priceConfig": self.pydantic_to_dict(price_config),
            "creditsConfig": self.pydantic_to_dict(credits_config),
            "nonce": nonce,
            "isTrialPlan": getattr(plan_metadata, "is_trial_plan", False),
        }

        options = self.get_backend_http_options("POST", body)
        url = f"{self.environment.backend}{API_URL_REGISTER_PLAN}"

        response = requests.post(url, **options)
        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to register plan. {response.status_code} - {response.text}"
            )

        return response.json()

    def register_credits_plan(
        self,
        plan_metadata: PlanMetadata,
        price_config: PlanPriceConfig,
        credits_config: PlanCreditsConfig,
    ) -> Dict[str, str]:
        """
        Allows an AI Builder to create a Payment Plan on Nevermined based on Credits.
        A Nevermined Credits Plan limits the access by the access/usage of the Plan.
        With them, AI Builders control the number of requests that can be made to an agent or service.
        Every time a user accesses any resource associated with the Payment Plan, the usage consumes from a capped amount of credits.
        When the user consumes all the credits, the plan automatically expires and the user needs to top up to continue using the service.

        Args:
            plan_metadata: Plan metadata
            price_config: Plan price configuration
            credits_config: Plan credits configuration

        Returns:
            The unique identifier of the plan (Plan ID) of the newly created plan

        Raises:
            PaymentsError: If the credits configuration is invalid
        """
        if credits_config.credits_type not in [
            PlanCreditsType.FIXED,
            PlanCreditsType.DYNAMIC,
        ]:
            raise PaymentsError.validation(
                "The creditsConfig.creditsType must be FIXED or DYNAMIC"
            )

        if credits_config.min_amount > credits_config.max_amount:
            raise PaymentsError.validation(
                "The creditsConfig.minAmount can not be more than creditsConfig.maxAmount"
            )

        return self.register_plan(plan_metadata, price_config, credits_config)

    def register_time_plan(
        self,
        plan_metadata: PlanMetadata,
        price_config: PlanPriceConfig,
        credits_config: PlanCreditsConfig,
    ) -> Dict[str, str]:
        """
        Allows an AI Builder to create a Payment Plan on Nevermined limited by duration.
        A Nevermined Credits Plan limits the access by the access/usage of the Plan.
        With them, AI Builders control the number of requests that can be made to an agent or service.
        Every time a user accesses any resource associated with the Payment Plan, the usage consumes from a capped amount of credits.
        When the user consumes all the credits, the plan automatically expires and the user needs to top up to continue using the service.

        Args:
            plan_metadata: Plan metadata
            price_config: Plan price configuration
            credits_config: Plan credits configuration

        Returns:
            The unique identifier of the plan (Plan ID) of the newly created plan

        Raises:
            PaymentsError: If the credits configuration is invalid
        """
        if credits_config.credits_type != PlanCreditsType.EXPIRABLE:
            raise PaymentsError.validation(
                "The creditsConfig.creditsType must be EXPIRABLE"
            )

        return self.register_plan(plan_metadata, price_config, credits_config)

    def register_credits_trial_plan(
        self,
        plan_metadata: PlanMetadata,
        price_config: PlanPriceConfig,
        credits_config: PlanCreditsConfig,
    ) -> Dict[str, str]:
        """
        Allows an AI Builder to create a Trial Payment Plan on Nevermined based on Credits.
        A Nevermined Trial Plan allow subscribers of that plan to test the Agents associated to it.
        A Trial plan is a plan that only can be purchased once by a user.

        Args:
            plan_metadata: Plan metadata
            price_config: Plan price configuration
            credits_config: Plan credits configuration

        Returns:
            The unique identifier of the plan (Plan ID) of the newly created plan
        """
        plan_metadata.is_trial_plan = True
        return self.register_credits_plan(plan_metadata, price_config, credits_config)

    def register_time_trial_plan(
        self,
        plan_metadata: PlanMetadata,
        price_config: PlanPriceConfig,
        credits_config: PlanCreditsConfig,
    ) -> Dict[str, str]:
        """
        Allows an AI Builder to create a Trial Payment Plan on Nevermined limited by duration.
        A Nevermined Trial Plan allow subscribers of that plan to test the Agents associated to it.
        A Trial plan is a plan that only can be purchased once by a user.

        Args:
            plan_metadata: Plan metadata
            price_config: Plan price configuration
            credits_config: Plan credits configuration

        Returns:
            The unique identifier of the plan (Plan ID) of the newly created plan
        """
        plan_metadata.is_trial_plan = True
        return self.register_time_plan(plan_metadata, price_config, credits_config)

    def get_plan(self, plan_id: str) -> Dict[str, Any]:
        """
        Get the metadata for a given Plan identifier.

        Args:
            plan_id: The unique identifier of the plan

        Returns:
            The plan's metadata

        Raises:
            PaymentsError: If the plan is not found
        """
        url = f"{self.environment.backend}{API_URL_GET_PLAN.format(plan_id=plan_id)}"
        response = requests.get(url)
        if not response.ok:
            raise PaymentsError.validation(
                f"Plan not found. {response.status_code} - {response.text}"
            )
        return response.json()

    def get_plan_balance(
        self, plan_id: str, account_address: Optional[str] = None
    ) -> Dict[str, Any]:
        """
        Get the balance of a plan for a specific account.

        Args:
            plan_id: The unique identifier of the plan
            account_address: The account address to check balance for (defaults to current user)

        Returns:
            The plan balance information

        Raises:
            PaymentsError: If unable to get plan balance
        """

        if not is_ethereum_address(account_address):
            account_address = self.get_account_address()

        url = f"{self.environment.backend}{API_URL_PLAN_BALANCE.format(plan_id=plan_id, holder_address=account_address)}"
        response = requests.get(
            url,
            headers={"Accept": "application/json", "Content-Type": "application/json"},
        )
        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to get plan balance. {response.status_code} - {response.text}"
            )
        return response.json()

    def order_plan(self, plan_id: str) -> Dict[str, bool]:
        """
        Order a plan by its ID.

        Args:
            plan_id: The ID of the plan to order

        Returns:
            The result of the order operation

        Raises:
            PaymentsError: If unable to order the plan
        """
        options = self.get_backend_http_options("POST")
        url = f"{self.environment.backend}{API_URL_ORDER_PLAN}".format(plan_id=plan_id)

        response = requests.post(url, **options)
        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to order plan. {response.status_code} - {response.text}"
            )
        return response.json()

    def mint_plan_credits(
        self, plan_id: str, credits_amount: int, credits_receiver: str
    ) -> Dict[str, Any]:
        """
        Mint credits for a plan.

        Args:
            plan_id: The unique identifier of the plan
            credits_amount: The amount of credits to mint
            credits_receiver: The address that will receive the credits

        Returns:
            The result of the mint operation

        Raises:
            PaymentsError: If unable to mint credits
        """
        body = {
            "planId": plan_id,
            "amount": credits_amount,
            "creditsReceiver": credits_receiver,
        }
        options = self.get_backend_http_options("POST", body)
        url = f"{self.environment.backend}{API_URL_MINT_PLAN}"

        response = requests.post(url, **options)
        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to mint credits. {response.status_code} - {response.text}"
            )
        return response.json()

    def mint_plan_expirable(
        self,
        plan_id: str,
        credits_amount: int,
        credits_receiver: str,
        credits_duration: int = 0,
    ) -> Dict[str, Any]:
        """
        Mint expirable credits for a plan.

        Args:
            plan_id: The unique identifier of the plan
            credits_amount: The amount of credits to mint
            credits_receiver: The address that will receive the credits
            credits_duration: The duration of the credits in seconds

        Returns:
            The result of the mint operation

        Raises:
            PaymentsError: If unable to mint credits
        """
        body = {
            "planId": plan_id,
            "creditsAmount": credits_amount,
            "creditsReceiver": credits_receiver,
            "creditsDuration": credits_duration,
        }
        options = self.get_backend_http_options("POST", body)
        url = f"{self.environment.backend}{API_URL_MINT_EXPIRABLE_PLAN}"

        response = requests.post(url, **options)
        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to mint expirable credits. {response.status_code} - {response.text}"
            )
        return response.json()

    def burn_credits(self, plan_id: str, credits_amount: str) -> Dict[str, Any]:
        """
        Burn credits from a plan.

        Args:
            plan_id: The unique identifier of the plan
            credits_amount: The amount of credits to burn

        Returns:
            The result of the burn operation

        Raises:
            PaymentsError: If unable to burn credits
        """
        body = {
            "planId": plan_id,
            "creditsAmount": credits_amount,
        }
        options = self.get_backend_http_options("POST", body)
        url = f"{self.environment.backend}{API_URL_BURN_PLAN}"

        response = requests.post(url, **options)
        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to burn credits. {response.status_code} - {response.text}"
            )
        return response.json()

    def get_agents_associated_to_plan(
        self, plan_id: str, pagination: Optional[PaginationOptions] = None
    ) -> Dict[str, Any]:
        """
        Gets the list of agents that can be accessed with a plan.

        Args:
            plan_id: The unique identifier of the plan
            pagination: Optional pagination options to control the number of results returned

        Returns:
            The list of all different agents giving access to the plan

        Raises:
            PaymentsError: If the plan is not found
        """
        if pagination is None:
            pagination = PaginationOptions()

        url = f"{self.environment.backend}{API_URL_GET_PLAN_AGENTS.format(plan_id=plan_id)}"
        params = {
            "page": pagination.page,
            "offset": pagination.offset,
        }
        response = requests.get(url, params=params)
        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to get agents associated to plan. {response.status_code} - {response.text}"
            )
        return response.json()

    # ------------------------------------------------------------------
    # Static helper methods to build price/credits configurations
    # These mirror functions in payments_py/plans.py for ergonomic access as
    # Payments.plans.<method>(...)
    # ------------------------------------------------------------------

    # Expose duration constants for convenience
    ONE_DAY_DURATION: int = plan_utils.ONE_DAY_DURATION
    ONE_WEEK_DURATION: int = plan_utils.ONE_WEEK_DURATION
    ONE_MONTH_DURATION: int = plan_utils.ONE_MONTH_DURATION
    ONE_YEAR_DURATION: int = plan_utils.ONE_YEAR_DURATION

    # Price configuration builders -------------------------------------
    @staticmethod
    def get_fiat_price_config(amount: int, receiver: str) -> PlanPriceConfig:
        """Build a fiat price configuration."""
        return plan_utils.get_fiat_price_config(amount, receiver)

    @staticmethod
    def get_crypto_price_config(
        amount: int,
        receiver: str,
        token_address: str = "0x0000000000000000000000000000000000000000",
    ) -> PlanPriceConfig:
        """Build a crypto (native/ERC20) price configuration."""
        return plan_utils.get_crypto_price_config(amount, receiver, token_address)

    @staticmethod
    def get_erc20_price_config(
        amount: int, token_address: str, receiver: str
    ) -> PlanPriceConfig:
        """Build an ERC20 price configuration."""
        return plan_utils.get_erc20_price_config(amount, token_address, receiver)

    @staticmethod
    def get_free_price_config() -> PlanPriceConfig:
        """Build a free price configuration."""
        return plan_utils.get_free_price_config()

    @staticmethod
    def get_native_token_price_config(amount: int, receiver: str) -> PlanPriceConfig:
        """Build a native token price configuration."""
        return plan_utils.get_native_token_price_config(amount, receiver)

    # Credits configuration builders -----------------------------------
    @staticmethod
    def get_expirable_duration_config(duration_of_plan: int) -> PlanCreditsConfig:
        """Build an expirable duration credits configuration."""
        return plan_utils.get_expirable_duration_config(duration_of_plan)

    @staticmethod
    def get_non_expirable_duration_config() -> PlanCreditsConfig:
        """Build a non-expirable duration credits configuration."""
        return plan_utils.get_non_expirable_duration_config()

    @staticmethod
    def get_fixed_credits_config(
        credits_granted: int, credits_per_request: int = 1
    ) -> PlanCreditsConfig:
        """Build a fixed credits configuration."""
        return plan_utils.get_fixed_credits_config(credits_granted, credits_per_request)

    @staticmethod
    def get_dynamic_credits_config(
        credits_granted: int,
        min_credits_per_request: int = 1,
        max_credits_per_request: int = 1,
    ) -> PlanCreditsConfig:
        """Build a dynamic credits configuration."""
        return plan_utils.get_dynamic_credits_config(
            credits_granted, min_credits_per_request, max_credits_per_request
        )

    @staticmethod
    def set_redemption_type(
        credits_config: PlanCreditsConfig, redemption_type: PlanRedemptionType
    ) -> PlanCreditsConfig:
        """Set redemption type on a credits configuration (returns new object)."""
        return plan_utils.set_redemption_type(credits_config, redemption_type)

    @staticmethod
    def set_proof_required(
        credits_config: PlanCreditsConfig, proof_required: bool = True
    ) -> PlanCreditsConfig:
        """Set proof requirement on a credits configuration (returns new object)."""
        return plan_utils.set_proof_required(credits_config, proof_required)

    def order_fiat_plan(self, plan_id: str) -> Dict[str, Any]:
        """
        Order a fiat plan using Stripe checkout.

        Args:
            plan_id: The unique identifier of the plan

        Returns:
            The Stripe checkout result

        Raises:
            PaymentsError: If unable to order the fiat plan
        """
        body = {"planId": plan_id, "sessionType": "embedded"}
        options = self.get_backend_http_options("POST", body)
        url = f"{self.environment.backend}{API_URL_STRIPE_CHECKOUT}"

        response = requests.post(url, **options)
        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to order fiat plan. {response.status_code} - {response.text}"
            )
        return response.json()

    def redeem_credits(
        self,
        agent_request_id: str,
        plan_id: str,
        redeem_from: str,
        credits_amount_to_redeem: str,
    ) -> Dict[str, Any]:
        """
        Redeem credits from a plan for a specific agent request.

        Args:
            agent_request_id: The unique identifier of the agent request
            plan_id: The unique identifier of the plan
            redeem_from: The address to redeem credits from
            credits_amount_to_redeem: The amount of credits to redeem

        Returns:
            The result of the redeem operation

        Raises:
            PaymentsError: If unable to redeem credits
        """
        body = {
            "agentRequestId": agent_request_id,
            "planId": plan_id,
            "redeemFrom": redeem_from,
            "amount": credits_amount_to_redeem,
        }
        options = self.get_backend_http_options("POST", body)
        url = f"{self.environment.backend}{API_URL_REDEEM_PLAN}"

        response = requests.post(url, **options)
        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to redeem credits. {response.status_code} - {response.text}"
            )
        return response.json()



"""
Error utilities and common JSON-RPC error codes used by the MCP paywall.
"""

from typing import Any, Dict


ERROR_CODES: Dict[str, int] = {
    "Misconfiguration": -32002,
    "PaymentRequired": -32003,
}


def create_rpc_error(code: int, message: str, data: Any | None = None) -> Exception:
    """
    Create an Exception that mimics JSON-RPC error objects with code and optional data.

     Args:
         code: Numeric error code.
         message: Human-readable error message.
         data: Optional structured payload to attach to the error.

     Returns:
         Exception configured with additional ``code`` and ``data`` attributes.
    """
    err = Exception(message)
    setattr(err, "code", code)
    if data is not None:
        setattr(err, "data", data)
    return err



"""
Authentication handler for MCP paywall.
"""

from typing import Any, Dict

from ..utils.request import extract_auth_header, strip_bearer
from ..utils.logical_url import build_logical_url, build_logical_meta_url
from ..utils.errors import create_rpc_error, ERROR_CODES


class PaywallAuthenticator:
    """
    Handles authentication and authorization for MCP requests using payments-py APIs.
    """

    def __init__(self, payments: Any) -> None:
        """Initialize the authenticator.

        Args:
            payments: Payments client used to call backend APIs.
        """
        self._payments = payments

    async def authenticate(
        self,
        extra: Any,
        options: Dict[str, Any],
        agent_id: str,
        server_name: str,
        name: str,
        kind: str,
        args_or_vars: Any,
    ) -> Dict[str, Any]:
        """Authenticate a tool/resource/prompt request.

        Args:
            extra: Extra request metadata containing headers.
            options: Paywall options used for the current handler.
            agent_id: Agent identifier configured in the server.
            server_name: Logical server name.
            name: Tool/resource/prompt name.
            kind: Handler kind (e.g. "tool", "resource", "prompt").
            args_or_vars: Arguments (tools/prompts) or variables (resources) for the request.

        Returns:
            A dictionary containing requestId, token, agentId and logicalUrl.

        Raises:
            Exception: When authorization is missing or the user is not a subscriber.
        """
        auth_header = extract_auth_header(extra)
        if not auth_header:
            raise create_rpc_error(
                ERROR_CODES["PaymentRequired"],
                "Authorization required",
                {"reason": "missing"},
            )
        token = strip_bearer(auth_header)
        logical_url = build_logical_url(
            {
                "kind": kind,
                "serverName": server_name,
                "name": name,
                "argsOrVars": args_or_vars,
            }
        )
        try:
            start = self._payments.requests.start_processing_request(
                agent_id,
                token,
                logical_url,
                "POST",
            )
            # support sync or async clients
            if hasattr(start, "__await__"):
                start = await start

            if not start or not start.get("balance", {}).get("isSubscriber", False):
                raise ValueError("Not a subscriber")

            return {
                "requestId": start.get("agentRequestId"),
                "token": token,
                "agentId": agent_id,
                "logicalUrl": logical_url,
            }
        except Exception:
            plans_msg = ""
            try:
                plans = self._payments.agents.get_agent_plans(agent_id)
                items = (plans or {}).get("plans", [])
                if isinstance(items, list) and items:
                    # Prefer human-readable names from metadata.main.name
                    names = []
                    for p in items:
                        meta_main = ((p or {}).get("metadata") or {}).get("main") or {}
                        pname = meta_main.get("name")
                        if isinstance(pname, str) and pname:
                            names.append(pname)
                    if names:
                        summary = ", ".join(names[:3])
                        plans_msg = f" Available plans: {summary}..."
            except Exception:
                pass

            raise create_rpc_error(
                ERROR_CODES["PaymentRequired"],
                f"Payment required.{plans_msg}",
                {"reason": "invalid"},
            )

    async def authenticate_meta(
        self, extra: Any, agent_id: str, server_name: str, method: str
    ):
        """Authenticate a meta operation (initialize/list/etc.).

        Args:
            extra: Extra request metadata containing headers.
            agent_id: Agent identifier configured in the server.
            server_name: Logical server name.
            method: Meta method name.

        Returns:
            A dictionary containing requestId, token, agentId and logicalUrl.

        Raises:
            Exception: When authorization is missing or the user is not a subscriber.
        """
        auth_header = extract_auth_header(extra)
        if not auth_header:
            raise create_rpc_error(
                ERROR_CODES["PaymentRequired"],
                "Authorization required",
                {"reason": "missing"},
            )
        token = strip_bearer(auth_header)
        logical_url = build_logical_meta_url(server_name, method)

        try:
            start = self._payments.requests.start_processing_request(
                agent_id,
                token,
                logical_url,
                "POST",
            )
            if hasattr(start, "__await__"):
                start = await start
            if not start or not start.get("balance", {}).get("isSubscriber", False):
                raise ValueError("Not a subscriber")
            return {
                "requestId": start.get("agentRequestId"),
                "token": token,
                "agentId": agent_id,
                "logicalUrl": logical_url,
            }
        except Exception:
            plans_msg = ""
            try:
                plans = self._payments.agents.get_agent_plans(agent_id)
                if hasattr(plans, "__await__"):
                    plans = await plans
                items = (plans or {}).get("plans", [])
                if isinstance(items, list) and items:
                    top = items[:3]
                    summary = ", ".join(
                        f"{p.get('planId') or p.get('id') or 'plan'}"
                        + (f" ({p.get('name')})" if p.get("name") else "")
                        for p in top
                    )
                    plans_msg = f" Available plans: {summary}..." if summary else ""
            except Exception:
                pass
            raise create_rpc_error(
                ERROR_CODES["PaymentRequired"],
                f"Payment required.{plans_msg}",
                {"reason": "invalid"},
            )



"""
MCP integration entry-point for the Nevermined Payments Python SDK.

This module exposes a class-based API (no dict-like compatibility):

- ``MCPIntegration``: Main integration surface with methods
  - ``configure(options)``: Set shared configuration such as agentId/serverName
  - ``with_paywall(handler, options)``: Decorate a handler with paywall
  - ``attach(server)``: Returns an object with ``registerTool``, ``registerResource``, ``registerPrompt``
  - ``authenticate_meta(extra, method)``: Authenticate meta operations like initialize/list

- ``build_mcp_integration(payments_service)``: Factory returning an ``MCPIntegration`` instance.
"""

from typing import Any, Awaitable, Callable, Dict, Protocol, Union

from .core.auth import PaywallAuthenticator
from .core.credits_context import CreditsContextProvider
from .core.paywall import PaywallDecorator
from .types import PaywallOptions, ToolOptions, PromptOptions, ResourceOptions


class _AttachableServer(Protocol):
    def registerTool(
        self, name: str, config: Any, handler: Callable[..., Awaitable[Any]]
    ) -> None:
        """Register a tool handler on the server."""
        pass

    def registerResource(
        self,
        name: str,
        template: Any,
        config: Any,
        handler: Callable[..., Awaitable[Any]],
    ) -> None:
        """Register a resource handler on the server."""
        pass

    def registerPrompt(
        self, name: str, config: Any, handler: Callable[..., Awaitable[Any]]
    ) -> None:
        """Register a prompt handler on the server."""
        pass


class MCPIntegration:
    """Class-based MCP integration for Payments.

    Provides a clean methods API to configure paywall, decorate handlers and attach
    registrations to a server implementation.
    """

    def __init__(self, payments_service: Any) -> None:
        """Initialize the integration with a Payments service instance.

        Args:
            payments_service: The initialized Payments client
        """
        self._payments = payments_service
        self._authenticator = PaywallAuthenticator(self._payments)
        self._credits_context = CreditsContextProvider()
        self._decorator = PaywallDecorator(
            self._payments, self._authenticator, self._credits_context
        )

    def configure(self, options: Dict[str, Any]) -> None:
        """Configure shared options such as ``agentId`` and ``serverName``.

        Args:
            options: Configuration dictionary with keys like ``agentId`` and ``serverName``
        """
        self._decorator.configure(options)

    def with_paywall(
        self,
        handler: Callable[..., Awaitable[Any]] | Callable[..., Any],
        options: Union[ToolOptions, PromptOptions, ResourceOptions, None] = None,
    ) -> Callable[..., Awaitable[Any]]:
        """Wrap a handler with the paywall protection.

        Args:
            handler: The tool/resource/prompt handler to protect
            options: The paywall options including kind, name and credits

        Returns:
            An awaitable handler with paywall applied
        """
        opts: PaywallOptions = options or {"kind": "tool", "name": "unnamed"}  # type: ignore[assignment]
        return self._decorator.protect(handler, opts)

    async def authenticate_meta(self, extra: Any, method: str) -> Dict[str, Any]:
        """Authenticate meta endpoints such as initialize/list.

        Args:
            extra: Extra request metadata containing headers
            method: The meta method name

        Returns:
            Authentication result dict
        """
        cfg: Dict[str, Any] = getattr(self._decorator, "config", {})  # type: ignore[assignment]
        agent_id = cfg.get("agentId", "")
        server_name = cfg.get("serverName", "mcp-server")
        return await self._authenticator.authenticate_meta(
            extra, agent_id, server_name, method
        )

    def attach(self, server: _AttachableServer):
        """Attach helpers to a server and return registration methods.

        Args:
            server: An object exposing registerTool/registerResource/registerPrompt

        Returns:
            An object with methods to register protected handlers on the server
        """

        integration = self

        class _Registrar:
            """Helper that registers protected handlers into the provided server."""

            def registerTool(
                self,
                name: str,
                config: Any,
                handler: Callable[..., Awaitable[Any]] | Callable[..., Any],
                options: Dict[str, Any] | None = None,
            ) -> None:
                """Register a tool handler protected by the paywall."""
                protected = integration.with_paywall(
                    handler, {"kind": "tool", "name": name, **(options or {})}
                )
                server.registerTool(name, config, protected)

            def registerResource(
                self,
                name: str,
                template: Any,
                config: Any,
                handler: Callable[..., Awaitable[Any]] | Callable[..., Any],
                options: Dict[str, Any] | None = None,
            ) -> None:
                """Register a resource handler protected by the paywall."""
                protected = integration.with_paywall(
                    handler, {"kind": "resource", "name": name, **(options or {})}
                )
                server.registerResource(name, template, config, protected)

            def registerPrompt(
                self,
                name: str,
                config: Any,
                handler: Callable[..., Awaitable[Any]] | Callable[..., Any],
                options: Dict[str, Any] | None = None,
            ) -> None:
                """Register a prompt handler protected by the paywall."""
                protected = integration.with_paywall(
                    handler, {"kind": "prompt", "name": name, **(options or {})}
                )
                server.registerPrompt(name, config, protected)

        return _Registrar()


def build_mcp_integration(payments_service: Any) -> MCPIntegration:
    """Factory that builds the class-based MCP integration.

    Args:
        payments_service: The initialized Payments client

    Returns:
        MCPIntegration instance
    """
    return MCPIntegration(payments_service)



"""High-level PaymentsClient wrapping the standard A2A JSON-RPC client with automatic auth."""

from __future__ import annotations

from typing import AsyncGenerator, Any

from a2a.client.client import ClientConfig
from a2a.client.client_factory import ClientFactory, minimal_agent_card
from a2a.client.middleware import ClientCallContext
from a2a.types import (
    MessageSendParams,
    TaskQueryParams,
    TaskPushNotificationConfig,
    TaskIdParams,
)

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from payments_py.payments import Payments


class PaymentsClient:  # noqa: D101
    def __init__(
        self, *, agent_base_url: str, payments: "Payments", agent_id: str, plan_id: str
    ) -> None:
        # Preserve trailing slash to avoid JSON-RPC 307 redirects between /a2a and /a2a/
        self._agent_base_url = (
            agent_base_url if agent_base_url.endswith("/") else agent_base_url + "/"
        )
        self._payments = payments
        self._agent_id = agent_id
        self._plan_id = plan_id
        self._access_token: str | None = None
        self._client = None  # Lazily created to ease unit testing

    # ------------------------------------------------------------------
    # Internal helpers
    # ------------------------------------------------------------------
    async def _get_access_token(self) -> str:
        if self._access_token is None:
            token_resp = await self._payments.agents.get_agent_access_token(
                self._plan_id, self._agent_id
            )
            self._access_token = token_resp.access_token  # type: ignore[attr-defined]
        return self._access_token

    def _auth_headers(self, token: str) -> dict[str, str]:
        return {"Authorization": f"Bearer {token}"}

    def _build_context(self, token: str) -> ClientCallContext:
        ctx = ClientCallContext()
        ctx.state["http_kwargs"] = {"headers": self._auth_headers(token)}
        return ctx

    def _call_with_optional_http_kwargs(self, method_name: str, *args, **kwargs):
        """Call client method trying http_kwargs first, then fallback without it.

        This keeps unit tests (that assert http_kwargs presence) satisfied while
        remaining compatible with a2a-sdk versions where http_kwargs is not an
        accepted parameter and headers are read from context.
        """
        client = self._get_client()
        method = getattr(client, method_name)
        # Try with http_kwargs if provided
        try:
            return method(*args, **kwargs)
        except TypeError:
            # Remove http_kwargs and retry
            if "http_kwargs" in kwargs:
                kwargs = {k: v for k, v in kwargs.items() if k != "http_kwargs"}
                return method(*args, **kwargs)
            raise

    @staticmethod
    def _extract_message(params: MessageSendParams | dict[str, Any]) -> Any:
        # Accept both pydantic model and plain dict
        if hasattr(params, "message"):
            return getattr(params, "message")
        return params.get("message")  # type: ignore[return-value]

    def _get_client(self):  # noqa: D401
        if self._client is None:
            # Ensure streaming enabled in config
            factory = ClientFactory(config=ClientConfig(streaming=True))
            client = factory.create(minimal_agent_card(self._agent_base_url))
            # Hint streaming support on the minimal card to allow resubscribe without fetching extended card
            try:
                if hasattr(client, "_card") and hasattr(client._card, "capabilities"):
                    setattr(client._card.capabilities, "streaming", True)
                # Avoid triggering authenticated extended card path in transports
                if hasattr(client._card, "supports_authenticated_extended_card"):
                    setattr(client._card, "supports_authenticated_extended_card", False)
            except Exception:
                pass
            self._client = client
        return self._client

    # ------------------------------------------------------------------
    # Public API
    # ------------------------------------------------------------------
    async def send_message(self, params: MessageSendParams) -> Any:  # noqa: D401
        token = await self._get_access_token()
        context = self._build_context(token)
        # BaseClient.send_message expects a Message, not MessageSendParams
        message_obj = self._extract_message(params)
        auth_headers = self._auth_headers(token)
        result = self._call_with_optional_http_kwargs(
            "send_message",
            message_obj,
            context=context,
            http_kwargs={"headers": auth_headers},
        )  # type: ignore[attr-defined]

        # Some client implementations return an async iterator (stream), others a coroutine
        stream_like = getattr(result, "__aiter__", None)
        if callable(stream_like):
            try:
                first_item = None
                async for item in result:  # type: ignore[func-returns-value]
                    first_item = item
                    break
                return first_item
            except StopAsyncIteration:  # pragma: no cover
                return None
        # Fallback: await coroutine and return its value
        return await result  # type: ignore[no-any-return]

    async def send_message_stream(
        self, params: MessageSendParams
    ) -> AsyncGenerator[Any, None]:  # noqa: D401
        token = await self._get_access_token()
        context = self._build_context(token)
        message_obj = self._extract_message(params)
        auth_headers = self._auth_headers(token)
        result = self._call_with_optional_http_kwargs(
            "send_message",
            message_obj,
            context=context,
            http_kwargs={"headers": auth_headers},
        )  # type: ignore[attr-defined]

        stream_like = getattr(result, "__aiter__", None)
        if callable(stream_like):
            async for item in result:  # type: ignore[func-returns-value]
                yield item
            return
        yield await result  # type: ignore[misc]

    async def get_task(self, params: TaskQueryParams) -> Any:  # noqa: D401
        token = await self._get_access_token()
        context = self._build_context(token)
        # type: ignore[arg-type]
        return await self._client.get_task(params, context=context)  # type: ignore[union-attr]

    async def set_task_push_notification_config(
        self, params: TaskPushNotificationConfig
    ) -> Any:  # noqa: D401
        token = await self._get_access_token()
        context = self._build_context(token)
        return await self._get_client().set_task_callback(params, context=context)  # type: ignore[union-attr, attr-defined]

    async def get_task_push_notification_config(
        self, params: TaskIdParams
    ) -> Any:  # noqa: D401
        token = await self._get_access_token()
        context = self._build_context(token)
        return await self._get_client().get_task_callback(params, context=context)  # type: ignore[union-attr, attr-defined]

    async def resubscribe_task(
        self, params: TaskIdParams
    ) -> AsyncGenerator[Any, None]:  # noqa: D401
        """Resubscribe to an existing task's event stream."""
        token = await self._get_access_token()
        context = self._build_context(token)
        # Ensure params conforms to TaskIdParams with required 'id'
        if isinstance(params, dict):
            task_id = params.get("taskId") or params.get("id")
        else:
            task_id = getattr(params, "task_id", None) or getattr(params, "id", None)
        normalized = {"id": task_id} if task_id else params  # type: ignore[assignment]

        auth_headers = self._auth_headers(token)
        res = self._call_with_optional_http_kwargs(
            "resubscribe",
            normalized,
            context=context,
            http_kwargs={"headers": auth_headers},
        )  # type: ignore[attr-defined]
        stream_like = getattr(res, "__aiter__", None)
        if callable(stream_like):
            async for item in res:  # type: ignore[func-returns-value]
                yield item
            return
        # If a coroutine is returned, await once and yield the value
        yield await res  # type: ignore[misc]

    # Utilities --------------------------------------------------------
    def clear_token(self) -> None:  # noqa: D401
        """Clear cached access-token forcing a refresh on next call."""
        self._access_token = None



"""
The AgentRequestsAPI class provides methods to manage the requests received by AI Agents integrated with Nevermined.
"""

import requests
from typing import Dict, Any
from payments_py.common.payments_error import PaymentsError
from payments_py.common.types import (
    PaymentOptions,
    TrackAgentSubTaskDto,
    StartAgentRequest,
)
from payments_py.api.base_payments import BasePaymentsAPI
from payments_py.api.nvm_api import (
    API_URL_REDEEM_PLAN,
    API_URL_INITIALIZE_AGENT,
    API_URL_TRACK_AGENT_SUB_TASK,
)
from payments_py.utils import decode_access_token


class AgentRequestsAPI(BasePaymentsAPI):
    """
    The AgentRequestsAPI class provides methods to manage the requests received by AI Agents integrated with Nevermined.
    """

    @classmethod
    def get_instance(cls, options: PaymentOptions) -> "AgentRequestsAPI":
        """
        Get a singleton instance of the AgentRequestsAPI class.

        Args:
            options: The options to initialize the payments class

        Returns:
            The instance of the AgentRequestsAPI class
        """
        return cls(options)

    def start_processing_request(
        self,
        agent_id: str,
        access_token: str,
        url_requested: str,
        http_method_requested: str,
    ) -> StartAgentRequest:
        """
        This method initializes an agent request.

        Args:
            agent_id: The unique identifier of the AI Agent
            access_token: The access token provided by the subscriber to validate
            url_requested: The URL requested by the subscriber to access the agent's API
            http_method_requested: The HTTP method requested by the subscriber to access the agent's API

        Returns:
            The information about the initialization of the request

        Raises:
            PaymentsError: If unable to initialize the agent request
        """
        initialize_agent_url = API_URL_INITIALIZE_AGENT.format(agent_id=agent_id)
        body = {
            "accessToken": access_token,
            "endpoint": url_requested,
            "httpVerb": http_method_requested,
        }
        options = self.get_backend_http_options("POST", body)

        url = f"{self.environment.backend}{initialize_agent_url}"
        response = requests.post(url, **options)
        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to validate access token. {response.status_code} - {response.text}"
            )

        return response.json()

    def redeem_credits_from_request(
        self,
        agent_request_id: str,
        request_access_token: str,
        credits_to_burn: int,
    ) -> Dict[str, Any]:
        """
        Allows the agent to redeem credits from a request.

        Args:
            agent_request_id: The unique identifier of the agent request
            request_access_token: The access token of the request
            credits_to_burn: The number of credits to burn

        Returns:
            A promise that resolves to the result of the operation

        Raises:
            PaymentsError: If unable to redeem credits from the request
        """
        # Decode the access token to get the wallet address and plan ID
        decoded_token = decode_access_token(request_access_token)
        if not decoded_token:
            raise PaymentsError.validation("Invalid access token provided")

        # Extract wallet address and plan ID from the token
        # Check if authToken is a nested JWT string that needs to be decoded
        auth_token_value = decoded_token.get("authToken")
        if auth_token_value and isinstance(auth_token_value, str):
            auth_token_decoded = decode_access_token(auth_token_value)
        else:
            auth_token_decoded = auth_token_value

        # Extract wallet address and plan ID with fallback logic like TypeScript version
        wallet_address = None
        if auth_token_decoded and isinstance(auth_token_decoded, dict):
            wallet_address = auth_token_decoded.get("sub")
        if not wallet_address:
            wallet_address = decoded_token.get("sub")

        plan_id = None
        if auth_token_decoded and isinstance(auth_token_decoded, dict):
            plan_id = auth_token_decoded.get("planId")
        if not plan_id:
            plan_id = decoded_token.get("planId")

        if not wallet_address or not plan_id:
            raise PaymentsError.validation(
                "Missing wallet address or plan ID in access token"
            )

        body = {
            "agentRequestId": agent_request_id,
            "planId": str(plan_id),  # Keep as string to avoid scientific notation
            "redeemFrom": wallet_address,
            "amount": credits_to_burn,
        }

        options = self.get_backend_http_options("POST", body)
        url = f"{self.environment.backend}{API_URL_REDEEM_PLAN}"
        response = requests.post(url, **options)
        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to redeem credits from request. {response.status_code} - {response.text}"
            )

        return response.json()

    def track_agent_sub_task(
        self, track_agent_sub_task: TrackAgentSubTaskDto
    ) -> Dict[str, Any]:
        """
        Tracks an agent sub task.

        This method is used by agent owners to track agent sub tasks for agent tasks.
        It records information about credit redemption, categorization tags, and processing descriptions.

        Args:
            track_agent_sub_task: The agent sub task data to track

        Returns:
            A promise that resolves to the result of the operation

        Raises:
            PaymentsError: If unable to track the agent sub task
        """
        body = {
            "agentRequestId": track_agent_sub_task.agent_request_id,
            "creditsToRedeem": track_agent_sub_task.credits_to_redeem or 0,
            "tag": track_agent_sub_task.tag,
            "description": track_agent_sub_task.description,
            "status": (
                track_agent_sub_task.status.value
                if track_agent_sub_task.status
                else None
            ),
        }

        options = self.get_backend_http_options("POST", body)
        url = f"{self.environment.backend}{API_URL_TRACK_AGENT_SUB_TASK}"
        response = requests.post(url, **options)

        if not response.ok:
            raise PaymentsError.internal(
                f"Unable to track agent sub task. {response.status_code} - {response.text}"
            )

        return response.json()
